/*
 * LM399_10V.c
 *
 * Created: 07.12.2018 20:40:08
 * Author : Andreas
 */ 
/*******************************************
Original code: adver (http://radiokot.ru/forum/viewtopic.php?p=3112844#p3112844)
Modified & ported to AVR Studio: Mickle T.
*******************************************
Compiler                : WinAVR
System Clock            : 5-20 MHz
Fuses					: XTAL, NO prescaler
*******************************************/
#define POWERSAVE 1
#define CLKTYPE 0
#if CLKTYPE == 2
#define F_CPU 4915200
#else 
  #if CLKTYPE == 1
#define F_CPU 8000000
  #else
#define F_CPU 20000000
  #endif
#endif
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
//******************************************
#define BASIS 14            // 14 bits of fractional part (8+14bit total => 2.4uV resolution)
#define DIVISOR (1<<14)
#define Key_Up()   0 // !(PINB & ~(1<<PB1))		// Is "UP" pressed?
#define Key_Down() 0 // !(PINB & ~(1<<PB2))		// Is "DOWN" pressed?
//******************************************
#define UZ  6.86477
#define UVal (UZ*4194304UL/10)  // 2879293
unsigned long Value = UVal;            
// Value have a fixed point format [8.14]
// 256/Value = amplification factor of 7-10 V converter
// If LM399 output voltage = U_Zener, then 
// Value = (U_Zener/10.000 V)*256*(1<<BASIS) = (U_Zener/10.000 V)*4194304
//******************************************
// MASH internal variables
unsigned int ValInt;                        // Integer part of Value
unsigned int ValFrac;                       // Fractional part of Value
unsigned int q[3]={1,0,0}; 	                // Latch outputs 16bit
char c[3][3]={{0,0,0}, {0,0,0}, {0,0,0}};   // Carry flip flops 1bit
unsigned int d[3];                          // Adder output / latch data input 16bit
//******************************************
// Calibration factor (Value), saved in EEPROM
EEMEM unsigned long CFG_CAL_VALUE = UVal;
//******************************************
void Init(void) 
{
    DDRB |= (1 << PB0);                 // PWM output
    DDRB &= ~((1<<PB1) | (1<<PB2));     // UP & DOWN key pins config
    PORTB |= (1 << PB1) | (1 << PB2);   // Pullup resistors

#if POWERSAVE == 1
    /* power saving modes */
	ACSR   = 0x80; /* switch off analog comparator */
	ADMUX  = 0x20; /* VREF = VCC, MUX = ADC0 (temperature sensor), ADC left aligned */
  #if CLKTYPE == 2
	ADCSRA = 0x05; /* ADC disabled clock = 4.9152 MHz / 2^5 */
  #else 
    #if CLKTYPE == 1
    ADCSRA = 0x06; /* ADC disabled clock = 8 MHz / 2^6 */
	#else
	ADCSRA = 0x07; /* ADC disabled clock = 20 MHz / 2^7 */
    #endif
  #endif
	DIDR0  = 0x20; /* disable ADC0 digital input buffer */
#endif
//    OSCCAL = 0x97-10; /* 0x97 = nominal - 5% */
	
	/* variable init */
    ValInt = Value >> BASIS;
    ValFrac = Value % DIVISOR;
    OCR0A = ValInt-1;          			// PWM compare match register initial value
    TCCR0A = (1<<COM0A1) | (1<<COM0A0) | (0<<COM0B1) | (0<<COM0B0) | (1<<WGM01) | (1<<WGM00);       
    // Set OC0A on Compare Match, BOTTOM, Normal mode, OC0B disconnected, Fast PWM Mode (WGM = 3) 
    TCCR0B = (0<<WGM02) | (0<<CS02) | (1<<CS01) | (0<<CS00);    // Prescaler CK/8 
    TIMSK = (1 << TOIE0);  				// Enable overflow interrupt
    sei();
}
//******************************************
ISR (TIM0_OVF_vect)
{
    // Carry flip-flops
    c[1][1] = c[1][0];  c[2][2] = c[2][1]; c[2][1] = c[2][0];
    // Adders
    d[0] = (q[0]+ValFrac) % DIVISOR;
    d[1] = (q[1]+d[0]) % DIVISOR;
    d[2] = (q[2]+d[1]) % DIVISOR;
    // Carries
    c[0][0] = (q[0]+ValFrac) >> BASIS;
    c[1][0] = (q[1]+d[0]) >> BASIS;
    c[2][0] = (q[2]+d[1]) >> BASIS;
    // Latch
    for (int i=0; i<3; i++) q[i] = d[i];
    OCR0A = ValInt - 1 
            + c[0][0]
            + c[1][0] -   c[1][1]
            + c[2][0] - 2*c[2][1] +   c[2][2];
}
//******************************************
// PWM configuration update
//****************************************** 
void Update(void) 
{
    cli();
    ValInt = Value >> BASIS;
    ValFrac = Value % DIVISOR;
    OCR0A = ValInt-1;          
    sei();
}
//******************************************
// Loading the calibration factor (Value) from EEPROM
//******************************************
void ConfigLoad (void)
{
// If the both keys was pressed on the startup,
// only the failsafe initial value is loaded !!!
    if (!Key_Up() && !Key_Down()) {
      eeprom_read_block(&Value,&CFG_CAL_VALUE,4);
//      _delay_ms(3000); 
    };
}
//******************************************
// Update calibration factor in EEPROM
//******************************************
void ConfigUpdate (void)
{
    eeprom_update_block(&Value,&CFG_CAL_VALUE,4);
    _delay_ms(3000);
}
//******************************************
int main( void ) 
{
    ConfigLoad();					// If UP & DOWN keys was pressed at startup
									// then load default config 
    Init();
#if POWERSAVE == 1	
    set_sleep_mode(SLEEP_MODE_IDLE);
#endif
    while(1) {
      if (Key_Up() && Key_Down()) 	// If UP & DOWN keys was pressed
	    ConfigUpdate(); 			// then save config to EEPROM
      if (Key_Up() && !Key_Down()) {
        Value ++;
        Update(); };
      if (!Key_Up() && Key_Down()) {
        Value --;
        Update(); };
//     _delay_ms(300);
#if POWERSAVE == 1
     sleep_mode();
#endif 
    };
}
